	page	132,63,1,1
	opt	rc
	title	'CPFSK Modulation/Demodulation'

;************************************************************
;* CPFSK.ASM -- CPFSK Modulation/Demodulation and	    *
;*		SSI drivers				    *
;*							    *
;* Provides interrupt based SSI handling, sample buffering  *
;* and CPFSK modulation and demodulation.		    *
;*							    *
;* CPFSK demodulation is based on article		    *
;*	Osborne, W., P., Lunz, M., B.:			    *
;*	"Coherent and Noncoherent Detection of CPFSK",	    *
;*	IEEE Trans. on Comm., Vol. 22, No. 8, August 1974   *
;* Symbol syncronization is based on article		    *
;*	Glisic, S., G.: 				    *
;*	"Symbol Syncronization in Digital Communications    *
;*	 Using Partial Response CPM signaling",             *
;*	IEEE Trans. on Comm., Vol. 37, No. 3, March 1989    *
;*							    *
;* This module uses registers as follows:		    *
;*  r3 - temporary use					    *
;*  r6 - input sample pointer				    *
;*  r7 - output sample pointer				    *
;*							    *
;* Copyright (C) 1990 by Jarkko Vuori. All rights reserved. *
;* Author(s): J Vuori					    *
;* Modification(s):					    *
;************************************************************

	section CPFSK
	xdef	ssi_ini,adc_ini
	xdef	ssi_rec,ssi_rer,ssi_xmt,ssi_xer
	xdef	xmiton,stxmit,tstcar
	xref	putbit,getbit

	org	p:

	nolist
	include 'ioequlc'
	include 'macros'
	include 'tlc32044'
	list

; CPFSK sync parameters
cM0	equ	 3				    ; random walk filter constant
adjlen	equ	23				    ; ADC timing control lenght (in samples)
analen	equ	28				    ; sample quality analyzing lenght
trshold equ	22				    ; sample quality decision treshold

; module status bits
mode	equ	0				    ; xmit/rec mode
phase	equ	1				    ; training/data phase
dcd	equ	2				    ; carrier on/off

; complex correlation (result in a1)
ccorr	macro
; s0 real -> b (real mode)
	conv	0,m*bits,r6,r3,a
	move	a1,x0				    ; square result
	mpy	x0,x0,a

; s0 imaginary -> a (real mode)
	conv	0,m*bits,r6,r3,b
	move	b1,x0				    ; square result
	macr	x0,x0,a
	endm


;****************************
;*    SSI initialization    *
;****************************
ssi_ini
; initialize SSI,
	IF !DEBUG
	movep		    #$4000,x:m_cra	    ; 16 bit word
	movep		    #$f200,x:m_crb	    ; syncronous,word frame,ext clk
	ELSE
	movep		    #$4008,x:m_cra	    ; 16 bit word
	movep		    #$f03c,x:m_crb	    ; asyncronous,word frame,ext clk
	ENDIF

; sample handling,
	move		    #costab,r7
	move		    #<ent-1,m7
	move		    #<f1,n7
	movi	<1,x:<xsamc

	move		    #<samples,r6
	move		    #<m*bits-1,m6
	movi	<m,x:<rsamc
	movi	0,x:<samque

; symbol syncro logic,
	movi	cM0,x:<rwalk
	movi	analen,x:<anacnt
	movi	0,x:<goodcnt

; mode flag, ADC command word and xmit waiting subroutine address
	movi	$0,x:<mode
	movi	0,x:<adcnt
	movi	0,x:<coffadr

	rts


;****************************
;*    ADC initialization    *
;****************************
adc_ini
; program the A/D & D/A converter chip (TLC32044, master clock is 5.184MHz)
; enable A/D high-pass filter, disable loopback, disable AUX IN,
; select syncronous mode, select input span 3Vpp, insert sinx/x correction
; lowpass cutoff frequency at	3420Hz -> TA  = 10 (SCF = 259kHz)
; sampling frequency is 	7200Hz -> TB  = 36
; unit sampling correction time  1.4uS -> TA' =  7
	pgmtlc	%10101001,10,7,36
	rts


;****************************
;*   SSI receive interrupt  *
;****************************
ssi_rec
; save registers to be used
	stoab	ssir

; check xmit/rec mode
	jclr	#mode,x:status,ssi_r0

; xmit mode, flush received character
	movep		    x:m_rx,x0
	jmp	<ssi_end

; receive mode, read received character
ssi_r0	move		    #<refs,r3
	movep		    x:m_rx,x:(r6)+

; correlate with s0 signal
	ccorr
	move	a1,x1

; correlate with s1 signal
	ccorr

; calculate result, limit to binary and queue received bits
	sub	x1,a	    x:<samque,b1
	rol	b
	move		    b1,x:<samque

; check if one symbol (m samples) received
	move		    #>1,x0
	move		    #>-1,x1
	move		    x:<rsamc,b
	sub	x0,b	    #>m,y0
	move		    b1,x:<rsamc
	jne	<ssi_end
	move		    y0,x:<rsamc

; yes, analyze eye pattern (= symbol quality)
; (by checking that samples nearby decision sample are equal)
; (...|X****X|X****X|X**...)
;   where X ignored samples
;	  * analyzed sample
;	  | symbol boundaries
	clr	b	    #>%011110000,y0
	move		    x:<samque,b1
	and	y0,b
	cmp	y0,b	    #>%000000000,y0
	jeq	<ssi_r0a
	cmp	y0,b
	jne	<ssi_r0b

ssi_r0a move		    x:<goodcnt,b	    ; good quality symbol
	add	x0,b
	move		    b1,x:<goodcnt

ssi_r0b move		    x:<anacnt,b 	    ; update analyses counter
	sub	x0,b	    #>analen,y0
	move		    b1,x:<anacnt
	jne	<ssi_rs
	move		    y0,x:<anacnt

	move		    x:<goodcnt,b	    ; make dcd decision
	move		    #>trshold,y0
	cmp	y0,b
	jhs	<ssi_r0c
	bclr	#dcd,x:<status			    ; dcd off
	jcc	<ssi_r0d
	dcdoff
	move		    x:<coffadr,b	    ; call specified routine if needed
	tst	b	    b1,r3
	jeq	<ssi_r0d
	clr	b
	move		    b,x:<coffadr
	jsr	(r3)
	jmp	<ssi_r0d
ssi_r0c bset	#dcd,x:<status			    ; dcd on
	jcs	<ssi_r0d
	dcdon

ssi_r0d clr	b
	move		    b,x:<goodcnt

; check if there are transitions (by analyzing the sample between decision samples)
; (...|00X00*|00X00*|00X...)
;   where X decision sample
;	  * analyzed sample
;	  | symbol boundaries
ssi_rs	clr	a	    #>1,x0
	move		    #>-1,x1
	jclr	#0,x:<samque,ssi_l1
	jset	#6,x:<samque,ssi_le
; 1x0
	move		    x0,a
	btst	#3,x:<samque
	tcc	x1,a
	jmp	<ssi_le
ssi_l1	jclr	#6,x:<samque,ssi_le
; 0x1
	move		    x1,a
	btst	#3,x:<samque
	tcc	x0,a

; symbol syncro
ssi_le	tst	a	    x:<rwalk,x0
	jeq	<ssi_r4
	add	x0,a	    #>2*cM0,x0
	jpl	<ssi_r2
; retard sampling
	move		    #cM0,a1
	movib		    retard,x:<adccmd
	movib		    adjlen,x:<adcnt
	jmp	<ssi_r3
;
ssi_r2	cmp	x0,a
	jle	<ssi_r3
; advance sampling
	move		    #cM0,a1
	movib		    advance,x:<adccmd
	movib		    adjlen,x:<adcnt
;
ssi_r3	move		    a1,x:<rwalk

; make symbol decision (with NRZ-S decoding)
ssi_r4	move		    x:<prvrsym,x0
	move		    x:<samque,a1
	eor	x0,a	    a1,x:<prvrsym
	not	a
	move		    a1,x:<tmp
	btst	#6,x:<tmp
; forward to the HDLC handler
	jsr	<putbit

; restore registers
ssi_end rclab	ssir
	rti


;****************************
;*  SSI receive with errors *
;*	  interrupt	    *
;****************************
ssi_rer movep		    x:m_sr,x:tmp	    ; clear ROE
	movep		    x:m_rx,x:tmp

	rti


;****************************
;*    SSI xmit interrupt    *
;****************************
ssi_xmt
; save registers to be used
	stoab	ssix

	clr	a
	jclr	#mode,x:status,xmt2

; xmit mode, check if we have sent a whole symbol
	move		    x:<xsamc,a
	move	#>1,x0
	sub	x0,a	    #>m,x0
	move		    a,x:<xsamc
	jne	<xmt1
	move		    x0,x:<xsamc

; yes, check on what phase we are
	jclr	#phase,x:status,xmt0

; training phase, send zero
	and	#$fe,ccr
	jmp	<morebit

; data phase, get a new bit to xmit
xmt0	jsr	<getbit
	jne	<morebit
; there is no more bits to send, switch to receiving mode
	bclr	#mode,x:<status
	pttoff
	IF DEBUG
	bug	0
	ENDIF
	jmp	<scx_end

; NRZ-S coding
morebit rol	a	    x:<prvxsym,x0
	not	a
	eor	x0,a	    #<f1,n7
	ror	a	    a1,x:<prvxsym

; determine frequency
	jcc	<xmt1
	move		    #<f0,n7
	nop

; send a new bit
xmt1	move		    x:(r7)+n7,x0	    ; adjust output level
	move		    x:<gain,x1
	mpyr	x0,x1,a     #>$fffc<<8,x0
xmt2	and	x0,a	    x:<adcnt,b
	tst	b	    #>1,y0		    ; check if we must adjust timing
	jeq	<xmt3
	sub	y0,b	    x:<adccmd,x0	    ; yes, add converter chip control cmds
	or	x0,a	    b1,x:<adcnt
xmt3	movep		    a1,x:m_tx

; restore registers
scx_end rclab	ssix
	rti


;****************************
;*   SSI xmit with errors   *
;*	  interrupt	    *
;****************************
ssi_xer
	movep		    x:m_sr,x:m_tx	    ; clear TUE

	rti


;****************************
;*     Turn xmitter on	    *
;****************************
xmiton	move		    #costab,r7
	bset	#phase,x:<status
	bset	#mode,x:<status
	ptton

	rts


;****************************
;*	Start xmitting	    *
;****************************
stxmit	bclr	#phase,x:<status

	rts


;****************************
;* Test if carrier detected *
;****************************
; x0 must contain address to be called when carrier turns off
; returns Carry set when Carrier detected
tstcar	jclr	#dcd,x:<status,caroff
	move		    x0,x:<coffadr	    ; call this address when carrier gets off
	ori	#$01,ccr
	rts

caroff	andi	#$fe,ccr
	rts


;****************************
;*	 DATA - AREA	    *
;****************************

	org	x:

status	ds	1				    ; xmit on/xmit off status
adccmd	ds	1				    ; ADC timing command
adcnt	ds	1				    ; ADC timing command counter
rsamc	ds	1				    ; receiver sample counter
xsamc	ds	1				    ; xmitter sample counter
samque	ds	1				    ; received sample queue
prvrsym ds	1				    ; previous received symbol
prvxsym ds	1				    ; previous xmitted symbol
rwalk	ds	1				    ; random walk filter counter
coffadr ds	1				    ; special routine address to be called when carrier turns off
anacnt	ds	1				    ; signal quality analyzing lenght counter
goodcnt ds	1				    ; good symbol counter

tmp	ds	1

	org	y:

; Complex reference signals
; m=6, bits=3
refs

; s0
; real
	dc  0.03750000,-0.01212019,-0.00815880,-0.00334936
	dc -0.03165111,0.00394950,0.05000000,-0.01710101
	dc -0.03830222,0.04330127,0.00868241,-0.04924039
	dc  0.02500000,0.04106969,-0.01099232,-0.01250000
	dc -0.00150768,-0.02856969
; imaginary
	dc  0.02165064,0.02599184,-0.00296956,0.01250000
	dc -0.00558094,-0.04514295,0.00000000,0.04698463
	dc -0.03213938,-0.02500000,0.04924039,-0.00868241
	dc -0.04330127,0.01915111,0.03020114,-0.00334936
	dc  0.00855050,-0.00249952

; s1
; real
	dc  0.03750000,-0.01212019,-0.00815880,-0.00334936
	dc -0.03165111,0.00394950,0.05000000,0.02500000
	dc -0.02500000,-0.05000000,-0.02500000,0.02500000
	dc  0.05000000,0.00394950,-0.03165111,-0.00334936
	dc -0.00815880,-0.01212019
; imaginary
	dc  0.02165064,0.02599184,-0.00296956,0.01250000
	dc -0.00558094,-0.04514295,0.00000000,0.04330127
	dc  0.04330127,0.00000000,-0.04330127,-0.04330127
	dc -0.00000000,0.04514295,0.00558094,-0.01250000
	dc  0.00296956,-0.02599184

	endsec


;****************************
;*    M*BITS MODULO DATA    *
;****************************
	section CPFSKData
	xdef	m,bits,samples
	xdef	ent,f0,f1,costab

	org	x:

; CPFSK demod parameters
m	equ	6				    ; samples per symbol
bits	equ	3				    ; number of observed bits
samples ds	m*bits
	ds	64-m*bits			    ; for modulo justification

pi	equ	3.141592654
ent	equ	36
f0	equ	11
f1	equ	6
freq	equ	2.0*pi/@cvf(ent)

; cosine lookuptable
costab
count	set	0
	dup	ent
	dc	0.999999*@cos(@cvf(count)*freq)
count	set	count+1
	endm

	endsec

	end
