	page	132,63,1,1
	opt	rc
	title	'PAM Modulation/Demodulation'

;************************************************************
;* PAM.ASM -- PAM Modulation/Demodulation and		    *
;*	      SSI drivers				    *
;*							    *
;* Provides interrupt based SSI handling, sample buffering  *
;* and adaptive PAM modulation and demodulation.	    *
;*							    *
;* PAM modulation and demodulation is based on some ideas   *
;* presented in book					    *
;*	Proakis, J., G.:				    *
;*	"Digital Communications",			    *
;*	McGraw-Hill, 1983				    *
;*							    *
;* Adaptive filters are based on article		    *
;*	Qureshi, S., U., H.:				    *
;*	"Adaptive Equalization",			    *
;*	Proc. of the IEEE, Vol. 73, No. 9, September 1985   *
;*							    *
;* This module uses registers as follows:		    *
;*  r3 - temporary use					    *
;*  r2 - filter coefficient pointer			    *
;*  r6 - input sample pointer				    *
;*  r7 - output sample pointer				    *
;*							    *
;* Copyright (C) 1990 by Jarkko Vuori. All rights reserved. *
;* Author(s): J Vuori					    *
;* Modification(s):					    *
;************************************************************

	section PAM
	xdef	ssi_ini,adc_ini
	xdef	ssi_rec,ssi_rer,ssi_xmt,ssi_xer
	xdef	xmiton,stxmit,tstcar
	xref	putbit,getbit

	org	p:

	nolist
	include 'ioequlc'
	include 'macros'
	include 'tlc32044'
	list

; Parameters
poly	equ	$8010				    ; G3RUH,K9NG scrambler polynomial (x^17 + x^12 + 1)
adjlen	equ	23				    ; ADC timing control lenght (in samples)

; module status bits
mode	equ	0				    ; xmit/rec mode
phase	equ	1				    ; training/data phase
dcd	equ	2				    ; carrier on/off
outtog	equ	3
intog	equ	4


;****************************
;*    SSI initialization    *
;****************************
ssi_ini
; initialize SSI,
	IF !DEBUG
	movep		    #$4000,x:m_cra	    ; 16 bit word
	movep		    #$f200,x:m_crb	    ; syncronous,word frame,ext clk
	ELSE
	movep		    #$4008,x:m_cra	    ; 16 bit word
	movep		    #$f03c,x:m_crb	    ; asyncronous,word frame,ext clk
	ENDIF

; sample handling,
	move		    #<samples,r6
	move		    #N-1,m6
	move		    #-2,n6		    ; adjustment for filtering
	move		    #<c,r2
	move		    #N-1,m2

	move		    #<outsamp,r7
	move		    #<coslen-1,m7

; mode flag, ADC command word and xmit waiting subroutine address
	movi	$0,x:<mode
	movi	0,x:<adcnt
	movi	0,x:<coffadr

	rts


;****************************
;*    ADC initialization    *
;****************************
adc_ini
; program the A/D & D/A converter chip (TLC32044, master clock is 5.184MHz)
; enable A/D high-pass filter, disable loopback, disable AUX IN,
; select syncronous mode, select input span 3Vpp, insert sinx/x correction
; lowpass cutoff frequency at	3420Hz -> TA  = 10 (SCF = 259kHz)
; sampling frequency is 	7200Hz -> TB  = 36
; unit sampling correction time  1.4uS -> TA' =  7
	pgmtlc	%10101001,10,7,36
	rts


;****************************
;*   SSI receive interrupt  *
;****************************
ssi_rec
; save registers to be used
	stoab	ssir

; check xmit/rec mode
	jclr	#mode,x:status,ssi_r0

; xmit mode, flush received character
	movep		    x:m_rx,x0
	jmp	<ssi_end

; receive mode, read received character
ssi_r0	movep		    x:m_rx,x:(r6)

; fractional spaced equalizer (FSE) - filter part
	conv		    0,N,r6,r2,a

; check if we must continue processing
	bchg	#intog,x:<status
	jcc	<ssi_end

; quantize and unscramble symbol, databit in N flag, result in C
	move		    x:<usrem,b1
	jmi	<unscram
	move		    #>(poly<<1)|1,x0
	eor	x0,b
unscram lsr	b
	move		    b1,x:<usrem

; make symbol decision (with NRZ-S decoding)
ssi_r4	rol	b	    x:<prvrsym,x0
	eor	x0,b	    b1,x:<prvrsym
	not	b
	lsr	b	    a,x:<tmp
; forward to the HDLC handler
	jsr	<putbit
	move		    x:<tmp,a

; calculate the error (desired - received) and multiply by alpha
	neg	a	    #-1.0,x1
	jpl	<in1
	move		    #0.9999999,x1
in1	add	x1,a	    #alpha,x0
	move	a,x1
	mpyr	x0,x1,a
	move	a,x1

; fractional Spaced Equalizer (FSE) - Adaptation part (LMS)
	move		    #<c,r3
	move		    x:(r6)+,x0	y:(r2)+,a   ; get x(0), c0
	do	#N,incupd
	macr	x0,x1,a     x:(r6)+,x0	y:(r2)+,y0  ; c = c + e(n) * x(n)
	tfr	y0,a			a,y:(r3)+   ; copy c, save new c
incupd
	move		    x:(r6)+n6,x0 y:(r2)-,y0

; restore registers
ssi_end rclab	ssir
	rti


;****************************
;*  SSI receive with errors *
;*	  interrupt	    *
;****************************
ssi_rer movep		    x:m_sr,x:tmp	    ; clear ROE
	movep		    x:m_rx,x:tmp

	rti


;****************************
;*    SSI xmit interrupt    *
;****************************
ssi_xmt
; save registers to be used
	stoab	ssix

	clr	a
	jclr	#mode,x:status,xmt2

; check if we need a new symbol
	bchg	#outtog,x:<status
	jcc	<ononeed

; yes, check on what phase we are
	jclr	#phase,x:status,xmt0

; training phase, send zero
	and	#$fe,ccr
	jmp	<morebit

; data phase, get a new bit to xmit
xmt0	jsr	<getbit
	jne	<morebit
; there is no more bits to send, switch to receiving mode
	bclr	#mode,x:<status
	pttoff
	jmp	<scx_end

; NRZ-S coding
morebit rol	a	    x:<prvxsym,x0
	not	a
	eor	x0,a
	ror	a	    a1,x:<prvxsym

; scrambler, databit in C flag, result in a (-1 or 1)
	clr	a
	rol	a	    x:<srem,x0
	eor	x0,a	    #-1.0,x1
	lsr	a	    #>poly,x0
	jcc	<scram
	eor	x0,a
scram	move		    a1,x:<srem
	move		    #0.9999999,a
	tcc	x1,a
	move		    a,x:<oprev
	jmp	<out1

ononeed move		    x:<oprev,a

; raised cosine shape output filter
out1	move		    #<rcos,r3
	move		    a,x:(r7)
	conv		    1,coslen,r7,r3,a

; send a new bit
	move		    a,x0		    ; adjust output level
	move		    x:<gain,x1
	mpyr	x0,x1,a     #>$fffc<<8,x0
xmt2	and	x0,a	    x:<adcnt,b
	tst	b	    #>1,y0		    ; check if we must adjust timing
	jeq	<xmt3
	sub	y0,b	    x:<adccmd,x0	    ; yes, add converter chip control cmds
	or	x0,a	    b1,x:<adcnt
xmt3	movep		    a1,x:m_tx

; restore registers
scx_end rclab	ssix
	rti


;****************************
;*   SSI xmit with errors   *
;*	  interrupt	    *
;****************************
ssi_xer
	movep		    x:m_sr,x:m_tx	    ; clear TUE

	rti


;****************************
;*     Turn xmitter on	    *
;****************************
xmiton	move		    #outsamp,r7
	bset	#phase,x:<status
	bset	#mode,x:<status
	ptton

	rts


;****************************
;*	Start xmitting	    *
;****************************
stxmit	bclr	#phase,x:<status

	rts


;****************************
;* Test if carrier detected *
;****************************
; x0 must contain address to be called when carrier turns off
; returns Carry set when Carrier detected
tstcar	jclr	#dcd,x:<status,caroff
	move		    x0,x:<coffadr	    ; call this address when carrier gets off
	ori	#$01,ccr
	rts

caroff	andi	#$fe,ccr
	rts


;****************************
;*	 DATA - AREA	    *
;****************************

	org	x:

status	ds	1				    ; xmit on/xmit off status
adccmd	ds	1				    ; ADC timing command
adcnt	ds	1				    ; ADC timing command counter
prvrsym ds	1				    ; previous received symbol
prvxsym ds	1				    ; previous xmitted symbol
coffadr ds	1				    ; special routine address to be called when carrier turns off
srem	ds	1
usrem	ds	1
oprev	ds	1
tmp	ds	1

	endsec


;****************************
;*    M*BITS MODULO DATA    *
;****************************
	section PAMData
	xdef	N,alpha,L,coslen
	xdef	outsamp,samples
	xdef	rcos,c

N	equ	11				    ; FSE lenght
alpha	equ	1.0/@cvf(N)			    ; LMS adaptation coefficient
L	equ	2				    ; samples/symbol
beta	equ	0.21				    ; rolloff parameter
coslen	equ	L*4*2				    ; raised cosine filter lenght

	org	x:

samples ds	N
	ds	16-N
outsamp ds	coslen

	org	y:

; raised cosine pulse filter constants
pi	equ	3.141592654
scaler	equ	2.0/(@cvf(coslen)/2.0)
rcos
count	set	0
	dup	coslen
t	set	(@cvf(count)-@cvf(coslen)/2.0)/@cvf(L)
	if	count-coslen/2
	dc	scaler*(@sin(pi*t)/(pi*t))*(@cos(beta*pi*t)/(1.0-4.0*beta*beta*t*t))
	else
	dc	scaler*0.9999999
	endif
count	set	count+1
	endm

; FSE filter constants
c	ds	N

	endsec

	end
